package uk.co.twey.util.compare;

import java.util.Comparator;
import java.lang.reflect.*;

public class ByField<ObjectT, FieldT extends Comparable<FieldT>> implements Comparator<ObjectT> {
    protected Class<ObjectT> objectT;
    protected Class<FieldT>  fieldT;
    protected String         getter;
    protected Object[]       args;
    protected Class<?>[]     argTs;

    public ByField(Class<ObjectT> objectT, Class<FieldT> fieldT, String getter, Object... args) {
        this.objectT = objectT;
        this.fieldT  = fieldT;
        this.getter  = getter;
        this.args    = args;
        this.argTs   = new Class<?>[args.length];

        for (int i = 0; i < args.length; ++i)
            argTs[i] = args[i].getClass();
    }

    public static <ObjectT, FieldT extends Comparable<FieldT>> ByField<ObjectT, FieldT> create(Class<ObjectT> objectT, Class<FieldT> fieldT, String getter, Object... args) {
        return new ByField<ObjectT, FieldT>(objectT, fieldT, getter, args);
    }

    public int compare(ObjectT o1, ObjectT o2) {
        try {
            Method m  = objectT.getDeclaredMethod(getter, argTs);
            Object v1 = m.invoke(o1, args),
                   v2 = m.invoke(o2, args);

            return (fieldT.cast(v1)).compareTo(fieldT.cast(v2));
        } catch (NoSuchMethodException e) {
            throw new NoSuchFieldException(objectT, getter, argTs);
        } catch (IllegalAccessException e) {
            throw new NoSuchFieldException(objectT, getter, argTs);
        } catch (ClassCastException e) {
            throw new IncorrectFieldTypeException(objectT, fieldT, getter, argTs);
        } catch (InvocationTargetException e) {
            // Impossible (?)
        }

        return 0;
    }

    protected static String methodToString(Class<?> objectT, String getter, Class<?>[] argTs) {
        StringBuilder str = new StringBuilder();

        str.append(objectT.getName());
        str.append(".");
        str.append(getter);
        str.append("(");

        for (int i = 0; i < argTs.length; ++i) {
            str.append(argTs[i].getName());
            if (i < argTs.length - 1)
                str.append(", ");
        }

        str.append(")");

        return str.toString();
    }

    public static class NoSuchFieldException extends RuntimeException {
        static final long serialVersionUID = -6280333810160217799L;

        private NoSuchFieldException(Class<?> objectT, String getter, Class<?>[] argTs) {
            super("Getter " + methodToString(objectT, getter, argTs) + " does not exist or is not accessible.");
        }
    }

    public static class IncorrectFieldTypeException extends RuntimeException {
        static final long serialVersionUID = 8387214789856909624L;

        private IncorrectFieldTypeException(Class<?> objectT, Class<?> fieldT, String getter, Class<?>[] argTs) {
            super("Getter " + methodToString(objectT, getter, argTs) + " does not return a " + fieldT.getName() + ".");
        }
    }
}
